MODULE Spline

  TYPE Spline1D
    DOUBLE PRECISION, ALLOCATABLE :: X(:), F(:), Coefs(:)
    INTEGER                       :: NX
  END TYPE Spline1D

  TYPE Spline2D
    DOUBLE PRECISION, ALLOCATABLE :: X(:), Y(:), F(:,:), CoefsX(:,:), CoefsY(:,:), CoefsXY(:,:)
    INTEGER                       :: NX, NY
  END TYPE

  INTERFACE Compute_Spline1D
    MODULE PROCEDURE Compute_Spline1D_Simple, Compute_Spline1D_Advanced
  END INTERFACE

CONTAINS

  SUBROUTINE Initialize_Spline_Module()
    IMPLICIT NONE

    COMMON /KODSPL/ kod, k, iw, iligne
    INTEGER :: kod, k, iw, iligne

    iw = 6
    iligne = 100
    k = 2
  END SUBROUTINE

  ! Allocate_Spline1D
  !   Allocate space to store a 1D spline with NX points
  SUBROUTINE Allocate_Spline1D( Spline, NX )
    IMPLICIT NONE

    TYPE (Spline1D) :: Spline
    INTEGER         :: NX

    Spline%NX = NX
    ALLOCATE( Spline%X(NX) )
    ALLOCATE( Spline%F(NX) )
    ALLOCATE( Spline%Coefs(NX) )
  END SUBROUTINE

  ! Allocate_Spline2D
  !   Allocate space to store a 2D spline with NX x NY points
  SUBROUTINE Allocate_Spline2D( Spline, NX, NY )
    IMPLICIT NONE

    TYPE (Spline2D) :: Spline
    INTEGER         :: NX, NY

    Spline%NX = NX
    Spline%NY = NY
    ALLOCATE( Spline%X(NX) )
    ALLOCATE( Spline%Y(NY) )
    ALLOCATE( Spline%F(NX,NY) )
    ALLOCATE( Spline%CoefsX(NX,NY) )
    ALLOCATE( Spline%CoefsY(NX,NY) )
    ALLOCATE( Spline%CoefsXY(NX,NY) )
  END SUBROUTINE

  ! Compute_Spline1D_Simple
  !   Compute a 1 dimensional spline without special boundary conditions
  SUBROUTINE Compute_Spline1D_Simple( Spline )
    IMPLICIT NONE

    TYPE (Spline1D)               :: Spline
    DOUBLE PRECISION, ALLOCATABLE :: alpha(:), beta(:), b(:)

    ALLOCATE( alpha( Spline%NX ), beta( Spline%NX ), b( Spline%NX ) )

    CALL DSPLIN( Spline%NX, Spline%X, Spline%F, Spline%Coefs, 0.0, 0, 0.0, 0, alpha, beta, b )

    DEALLOCATE( alpha, beta, b )
  END SUBROUTINE

  ! Compute_Spline1D_Advanced
  !   Compute a 1 dimensional spline with boundary conditions
  SUBROUTINE Compute_Spline1D_Advanced( Spline, Condition1, Der1, ConditionN, DerN )
    IMPLICIT NONE

    TYPE (Spline1D)  :: Spline
    INTEGER          :: Condition1, ConditionN
    DOUBLE PRECISION :: Der1, DerN
    DOUBLE PRECISION, ALLOCATABLE :: alpha(:), beta(:), b(:)

    ALLOCATE( alpha( Spline%NX ), beta( Spline%NX ), b( Spline%NX ) )

    CALL DSPLIN( Spline%NX, Spline%X, Spline%F, Spline%Coefs, Der1, Condition1, DerN, ConditionN, alpha, beta, b )

    DEALLOCATE( alpha, beta, b )
  END SUBROUTINE

  ! Evaluate_Spline1D
  !   Evaluate a 1 dimensional spline
  FUNCTION Evaluate_Spline1D( Spline, X ) RESULT ( Val )
    IMPLICIT NONE

    TYPE (Spline1D) :: Spline
    DOUBLE PRECISION :: X, Val

    Val = DPL( Spline%NX, Spline%X, Spline%F, Spline%Coefs, X )
  END FUNCTION

  FUNCTION Evaluate_Spline1D_Derivative( Spline, X ) RESULT ( Val )
    IMPLICIT NONE

    TYPE (Spline1D) :: Spline
    DOUBLE PRECISION :: X, Val

    Val = DPLP( Spline%NX, Spline%X, Spline%F, Spline%Coefs, X )
  END FUNCTION

  ! Compute_Spline2D
  !   Compute a 2 dimensional spline
  SUBROUTINE Compute_Spline2D( Spline )
    IMPLICIT NONE

    TYPE (Spline2D) :: Spline

    CALL SPLIN2D( Spline%X, Spline%NX, Spline%Y, Spline%NY, &
                  Spline%F, Spline%CoefsX, Spline%CoefsY, Spline%CoefsXY, Spline%NX, Spline%NY)
  END SUBROUTINE

  ! Evaluate_Spline2D
  !   Evaluate a 2 dimensional spline
  FUNCTION Evaluate_Spline2D( Spline, X, Y ) RESULT ( Val )
    IMPLICIT NONE

    TYPE (Spline2D)  :: Spline
    DOUBLE PRECISION :: X, Y, Val
    INTEGER :: i, j

    CALL DPL2D( X, i, Y, j, Spline%X, Spline%NX, Spline%Y, Spline%NY, &
                Spline%F, Spline%CoefsX, Spline%CoefsY, Spline%CoefsXY, Spline%NX, Spline%NY, Val )
  END FUNCTION

  SUBROUTINE Evaluate_Spline2D_Alt( Spline, X, Y, MyOut, switch )
    IMPLICIT NONE

    TYPE (Spline2D)  :: Spline
    INTEGER          :: iX, iY
    DOUBLE PRECISION :: X, Y, MyOut(3)
    LOGICAL          :: switch

    CALL UBICAR( Spline%X, X, Spline%NX, iX, switch )
    IF (switch) RETURN
    CALL UBICAR( Spline%Y, Y, Spline%NY, iY, switch )
    IF (switch) RETURN
    CALL SPLCOFII( Spline%X, Spline%Y, X, Y, iX, iY, Spline%NX, Spline%NY )
    CALL SPLINII( Spline%F, Spline%CoefsX, Spline%CoefsY, Spline%CoefsXY, iX, iY, MyOut, Spline%NX, Spline%NY )
  END SUBROUTINE

!**************************** D P L 2 D ******************************
      SUBROUTINE DPL2D(ax,im,ay,jn,x,m,y,n,f,cx,cy,cint,idx,idy,res)
!
!     Calculate f(x,y) in 2D interpolation.
!
!       Author: A. Salin     Version: 2.0    15/01/99

!     AX, AY: point for which the interpolate is required.
!     IM, JN: localization of AX and AY in the last call
!     X,M: set of m abcissae
!     Y,N: set of n ordinates
!     F(i,j): value of functions at (x_i,y_j)
!     CX, CY, CINT: arrays of dimension (idx,idy)
!                   determined by subroutine SPL2D or SPLCYC2D
!     IDX, IDY: dimensions in calling program.
!     RES : value of interpolate.

!     Required subroutine: LOCATE

      IMPLICIT REAL*8 (a-h,o-z)
      PARAMETER (s6=1.D0/6.D0)

      DIMENSION x(idx),y(idy),f(idx,idy),cx(idx,idy),cy(idx,idy),cint(idx,idy)

      CALL LOCATE(ax,im,x,m)
      CALL LOCATE(ay,jn,y,n)

      IF(im.LE.1.OR.im.GT.m) THEN
        WRITE(*,*) 'x=',ax,'outside interval in DPL2D'
        STOP
      END IF
      IF(jn.LE.1.OR.jn.GT.n) THEN
        WRITE(*,*) 'y=',ay,'outside interval in DPL2D'
        STOP
      END IF

      fx=x(im)-ax
      gx=ax-x(im-1)
      ex=fx+gx
      den=1.D0/ex
      c2=-gx*fx*s6*den
      c1=c2*(fx+ex)
      c2=c2*(gx+ex)
      c3=fx*den
      c4=gx*den
!                                              Determine F(x,y_(jn-1))
      cx1=cx(im-1,jn-1)
      cx2=cx(im,jn-1)
      af1=f(im-1,jn-1)
      af2=f(im,jn-1)
      afj=c1*cx1+c2*cx2+c3*af1+c4*af2
!                                               Determine F(x,y_(jn))
      cx1=cx(im-1,jn)
      cx2=cx(im,jn)
      af1=f(im-1,jn)
      af2=f(im,jn)
      afjp=c1*cx1+c2*cx2+c3*af1+c4*af2
!                                               Determine c_(jn-1)(x)
      cx1=cint(im-1,jn-1)
      cx2=cint(im,jn-1)
      af1=cy(im-1,jn-1)
      af2=cy(im,jn-1)
      cy1=c1*cx1+c2*cx2+c3*af1+c4*af2

!                                                 Determine c_(jn)(x)
      cx1=cint(im-1,jn)
      cx2=cint(im,jn)
      af1=cy(im-1,jn)
      af2=cy(im,jn)
      cy2=c1*cx1+c2*cx2+c3*af1+c4*af2
!                                                    Determine f(x,y)
      fy=y(jn)-ay
      gy=ay-y(jn-1)
      ey=fy+gy
      res=(-gy*fy*s6*(cy1*(fy+ey)+cy2*(gy+ey))+(gy*afjp+fy*afj))/ey

      END SUBROUTINE
!**************************** D P L C O F T **************************
      SUBROUTINE DPLCOFT(n,x,t,c,dc,k)

!     Calculates coefficients for the spline interpolation of a
!     function (C) and its derivative (DC) at point T. The N nodes are
!     X. K is such that t.GE.x(k-1) and t.LT.x(k).

      IMPLICIT REAL*8 (a-h,o-z)

      DIMENSION x(n),c(4),dc(4)

      IF(k.LE.1.OR.k.GT.n) THEN
        klo=0
        khi=n+1
      ELSE

      inc=1
      klo=k-1
      IF(t.GT.x(klo)) THEN
 10     khi=klo+inc
        IF(khi.GT.n) THEN
          khi=n+1
        ELSE IF(t.GT.x(khi)) THEN
          klo=khi
          inc=inc+inc
          GO TO 10
        END IF
      ELSE
        khi=klo
 20     klo=khi-inc
        IF(klo.LT.1) THEN
          klo=0
        ELSE IF(t.LT.x(klo)) THEN
          khi=klo
          inc=inc+inc
          GO TO 20
        END IF
      END IF

      END IF

100   CONTINUE
      IF(khi-klo.GT.1) THEN
        km=(khi+klo)/2
        IF(t.GT.x(km)) THEN
          klo=km
        ELSE
          khi=km
        END IF
        GO TO 100
      END IF
        k=klo+1

      IF(t.EQ.x(1)) k=2

        f=x(k)-t
        g=t-x(k-1)
        e=f+g
        se=1.D0/e
        se6=se/6.D0
        h=-g*f*se6
        c(1)=h*(f+e)
        c(2)=h*(g+e)
        c(3)=f*se
        c(4)=g*se
        dc(1)=-h-(f+e)*(f-g)*se6
        dc(2)=h-(g+e)*(f-g)*se6
        dc(3)=-se
        dc(4)=se

      END SUBROUTINE
!**************************** D P L *********************************
      FUNCTION DPL(n,x,y,m,t)

!        Version 3.0  - 15/1/97       Author: A. SALIN

!        Calculation of a function by interpolation. Double precision.
!        See comments in SUBROUTINE DSPLIN

      IMPLICIT REAL*8 (a-h,o-z)

      DIMENSION x(n),y(n)
      REAL*8 m(n)
      LOGICAL order

      COMMON/KODSPL/kod,k,iw,iligne

    1 FORMAT(' DPL: extrapolation - ',1PE15.8,' <',E15.8)
    2 FORMAT(' DPL: extrapolation - ',1PE15.8,' >',E15.8)

      order=x(2).GT.x(1)
      IF(k.LE.1.OR.k.GT.n) THEN
        klo=0
        khi=n+1
        GO TO 100
      END IF

      inc=1
      klo=k-1
      IF(t.GT.x(klo).EQV.order) THEN
 10     khi=klo+inc
        IF(khi.GT.n) THEN
          khi=n+1
        ELSE IF(t.GT.x(khi).EQV.order) THEN
          klo=khi
          inc=inc+inc
          GO TO 10
        END IF
      ELSE
        khi=klo
 20     klo=khi-inc
        IF(klo.LT.1) THEN
          klo=0
        ELSE IF(t.LT.x(klo).EQV.order) THEN
          khi=klo
          inc=inc+inc
          GO TO 20
        END IF
      END IF

 100  CONTINUE
      IF(khi-klo.GT.1) THEN
        km=(khi+klo)/2
        IF(t.GT.x(km).EQV.order) THEN
          klo=km
        ELSE
          khi=km
        END IF
        GO TO 100
      END IF
        k=klo+1

      IF(t.EQ.x(1)) k=2
      IF(k.LE.1) THEN
        e=x(2)-x(1)
        DPL=((y(2)-y(1))/e-m(2)*e/6.D0)*(t-x(1))+y(1)
        kod=1
        IF(iw.GT.0.AND.iligne.GT.0) THEN
          IF(order) THEN
            WRITE(iw,1) t,x(1)
          ELSE
            WRITE(iw,2) t,x(1)
          END IF
          iligne=iligne-1
        END IF
      ELSE IF(k.GT.n) THEN
        e=x(n)-x(n-1)
        DPL=((y(n)-y(n-1))/e+m(n-1)*e/6.D0)*(t-x(n))+y(n)
        kod=2
        IF(iw.GT.0.AND.iligne.GT.0) THEN
          IF(order) THEN
            WRITE(iw,2) t,x(n)
          ELSE
            WRITE(iw,1) t,x(n)
          END IF
          iligne=iligne-1
        END IF
      ELSE
        f=x(k)-t
        g=t-x(k-1)
        e=f+g
        DPL=(-g*f*(m(k-1)*(f+e)+m(k)*(g+e))+6.D0*(g*y(k)+f*y(k-1)))/(6.D0*e)
        kod=0
      END IF
      END FUNCTION
!**************************** D P L P *******************************
      FUNCTION DPLP(n,x,y,m,t)

!        Version 3.0  - 17/1/97       Author: A. SALIN

!        Calculation of the derivative of a function by interpolation.
!        Double precision. See comments in SUBROUTINE DSPLIN.
!        If t is outside the interval over which the function is
!        defined, the derivative is extrapolated linearly from its
!        value for the first two (last two) points of the interval.

      IMPLICIT REAL*8 (a-h,o-z)

      DIMENSION x(n),y(n)
      REAL*8 m(n)
      LOGICAL order

      COMMON/KODSPL/kod,k,iw,iligne

    1 FORMAT(' DPLP: extrapolation - ',1PE15.8,' <',E15.8)
    2 FORMAT(' DPLP: extrapolation - ',1PE15.8,' >',E15.8)

      order=x(2).GT.x(1)
      IF(k.LE.1.OR.k.GT.n) THEN
        klo=0
        khi=n+1
        GO TO 100
      END IF

      inc=1
      klo=k-1
      IF(t.GT.x(klo).EQV.order) THEN
 10     khi=klo+inc
        IF(khi.GT.n) THEN
          khi=n+1
        ELSE IF(t.GT.x(khi).EQV.order) THEN
          klo=khi
          inc=inc+inc
          GO TO 10
        END IF
      ELSE
        khi=klo
 20     klo=khi-inc
        IF(klo.LT.1) THEN
          klo=0
        ELSE IF(t.LT.x(klo).EQV.order) THEN
          khi=klo
          inc=inc+inc
          GO TO 20
        END IF
      END IF

 100  CONTINUE
      IF(khi-klo.GT.1) THEN
        km=(khi+klo)/2
        IF(t.GT.x(km).EQV.order) THEN
          klo=km
        ELSE
          khi=km
        END IF
        GO TO 100
      END IF
        k=klo+1

      IF(t.EQ.x(1)) k=2
      IF(k.LE.1) THEN
        e=x(2)-x(1)
        g=t-x(1)
        f=x(2)-t
        DPLP=(y(2)-y(1))/e+(m(2)*(2*g-f)+m(1)*(g-2*f))/6.D0
        kod=1
        IF(iw.GT.0.AND.iligne.GT.0) THEN
          IF(order) THEN
            WRITE(iw,1) t,x(1)
          ELSE
            WRITE(iw,2) t,x(1)
          END IF
          iligne=iligne-1
        END IF
      ELSE IF(k.GT.n) THEN
        e=x(n)-x(n-1)
        g=t-x(n-1)
        f=x(n)-t
        DPLP=(y(n)-y(n-1))/e+(m(n)*(2*g-f)+m(n-1)*(g-2*f))/6.D0
        kod=2
        IF(iw.GT.0.AND.iligne.GT.0) THEN
          IF(order) THEN
            WRITE(iw,2) t,x(n)
          ELSE
            WRITE(iw,1) t,x(n)
          END IF
          iligne=iligne-1
        END IF
      ELSE
        f=x(k)-t
        g=t-x(k-1)
        e=f+g
        e2=e*e
        DPLP=(-m(k-1)*(3.D0*f*f-e2)+m(k)*(3.D0*g*g-e2)+6.D0*(y(k)-y(k-1)))/(6.D0*e)
        kod=0
      END IF
      END FUNCTION
!**************************** D S P L I N ***************************
!        Interpolation package. Given a set of n couples x(i),y(i),
!     DSPLIN defines a cubic spline function F(x) such that:
!        1)- F(x(i)) = y(i)
!        2)- F(x), F'(x), F''(x) are continuous in the interval
!     [x(1),x(n)].
!        Required subroutine: TRIDIA.

!********************************************************************
!
      SUBROUTINE DSPLIN(n,x,y,cm,cm1,ic1,cmn,icn,alpha,beta,b)
!
!        Author: A. Salin   Version 3.2  20/1/86 - 10/6/97
!
!     ****SUBROUTINE DSPLIN: calculates the vector F''(x(i)) which
!     defines the cubic spline function F(x).
!          N= number of pivots x(i).
!          X= array of abcissae. Should be stored by increasing or
!     decreasing order.
!          Y= array of values of Y(i).
!          CM= array of F''(x(i)) with dimension equal to that of X
!     and Y.
!          IC1,CM1,ICN,CMN: define the condition at x(1) and x(n)
!     If IC1=0, function is the same in first two intervals.
!     If IC1=1, second derivative at x(1) given in CM1
!     If IC1=2, first derivative at x(1) given in CM1
!     IF IC1>2, zero second derivative at x(1)
!          Similar definitions for ICN and CMN around x(n)
!          ALPHA, BETA, B are working arrays of dimension N at least.
!     For cyclic spline use the program TRISPL.
!
!     ****DPLCOF: the interpolated function at R is:
!       F=C(1)*CM(K-1)+C(2)*CM(K)+C(3)*Y(K-1)+C(4)*Y(K)
!       where C is obtained from:
!       CALL DPLCOF(n,x,r,c,k)
!       The vector C should be of dimension 4.
!
!     ****Functions DPL,DPLP,DPLP2: calculate respectively the function
!     F(x), its first or second derivative:
!             N= number of pivots
!             X,Y,M= same as X,Y,CM in DSPLIN.
!             T= value for which the function (or its derivative)
!     must be determined.
!     WARNING: when T is outside the interval [X(1),X(N)], F(X)
!     *******  (resp. F', F") is determined by linear extrapolation
!              using the value of F (resp. F', F") for the first two
!              (last two) points in the interval.
!
!     ****Subroutine DINITI et DINTSP: calculate the integral of the
!     function F(x) from X(1) to T. The subroutine DINITI should first
!     be called before the first CALL DINTSP concerning a given
!     function F(x). DINITI calculates the array (CI) of values of the
!     integral of F(x) from X(1) to all X(i).
!
!        Arguments of DINITI:
!             N,X,Y,CM: as in subroutine DSPLIN.
!             CI: array of dimension at least N.
!        Arguments of DINTSP:
!             N,X,Y,CM,CI: as in DINITI
!             T: value of the upper limit of the integration (X(1).LT.T.
!     LE.X(N)).
!             B: value of the integral.
!       The calculations in DINTSP are nearly as rapid as for the deter
!     mination of the function F(x) by DPL.
!
!     Parameters of COMMON/KODSPL/:
!     -----------------------------------------------------
!       -KOD= after execution of one subroutine of the package, KOD
!     takes the value:
!              *0: no error
!              *-1: values of X(i) are not stored in correct order.
!              *1: T outside the interval [X(1),X(N)]. Linear extrapo-
!     lation done using X(1) and X(2).
!              *2: T outside the interval [X(1),X(N)]. Linear extrapo-
!     lation done using X(n-1) and X(n).
!       -K= after the execution of DPL, DPLP or DPLP2, K is such that
!     T lies in the interval [X(k-1),X(k)].
!       -IW= error messages are printed on unit IW unless IW.LE.0
!     Default: 6.
!       -ILIGNE: when T is outside the interval [X(1),X(n)], DPL, DPLP
!     and DPLP2 print a message if ILIGNE.GT.0. The initial value of
!     ILIGNE (defined by DATA) is 100. For every extrapolation, the
!     value is decreased by 1.
!
!     ***************************************************************

      IMPLICIT REAL*8 (a-h,o-z)
      PARAMETER (six=6.D0,douze=12.D0,zero=0.D0)

      DIMENSION x(n),y(n),cm(n)
      DIMENSION alpha(n),beta(n),b(n)

      COMMON/KODSPL/kod,k,iw,iligne

   91 FORMAT('Error in the data for DSPLIN',/,'Check abcissae:',1PD15.8,1X,' and ',D15.8,/,'Program stopped')

!                                           Contour condition at X(1)
      IF(ic1.EQ.0) THEN
        s1=zero
        fac1=(x(2)-x(1))*(1.D0+(x(2)-x(1))/(x(3)-x(2)))/six
        fab1=-(x(2)-x(1))**2/(six*(x(3)-x(2)))
      ELSE IF(ic1.EQ.1) THEN
        s1=-cm1*(x(2)-x(1))/six
        fac1=zero
        fab1=zero
      ELSE IF(ic1.EQ.2) THEN
        s1=0.5D0*(cm1-(y(2)-y(1))/(x(2)-x(1)))
        fac1=-(x(2)-x(1))/douze
        fab1=zero
      ELSE
!        WRITE(iw,*) 'DSPLIN: unknown condition - use default'
        s1=zero
        fac1=zero
        fab1=zero
      END IF
!                                           Contour condition at X(n)
      IF(icn.EQ.0) THEN
        sn=zero
        facn=(x(n)-x(n-1))*(1.D0+(x(n)-x(n-1))/(x(n-1)-x(n-2)))/six
        fabn=-(x(n)-x(n-1))**2/(six*(x(n-1)-x(n-2)))
      ELSE IF(icn.EQ.1) THEN
        sn=-cmn*(x(n)-x(n-1))/six
        facn=zero
        fabn=zero
      ELSE IF(icn.EQ.2) THEN
        sn=-0.5D0*(cmn-(y(n)-y(n-1))/(x(n)-x(n-1)))
        facn=-(x(n)-x(n-1))/douze
        fabn=zero
      ELSE
!        WRITE(iw,*) 'DSPLIN: unknown condition - use default'
        sn=zero
        facn=zero
        fabn=zero
      END IF

      kod=0
      c=x(2)-x(1)
      e=(y(2)-y(1))/c
      DO i=3,n
        i2=i-2
        a=c
        c=x(i)-x(i-1)
        IF(a*c.LE.zero) THEN
          kod=-1
          IF(iw.GT.0) WRITE(iw,91) x(i-1),x(i)
          STOP
          END IF
        alpha(i2)=(a+c)/3.D0
        beta(i2)=c/six
        cm(i2)=beta(i2)
        d=e
        e=(y(i)-y(i-1))/c
        b(i2)=e-d
      END DO

      b(1)=b(1)+s1
      b(n-2)=b(n-2)+sn
      alpha(1)=alpha(1)+fac1
      alpha(n-2)=alpha(n-2)+facn
      cm(n-3)=cm(n-3)+fabn
      beta(1)=beta(1)+fab1
!                                            Solve tridiagonal system
      CALL TRIDIA(alpha,cm,beta,b,cm(2),n-2)

      IF(ic1.EQ.0) THEN
        cm(1)=cm(2)*(1.D0+(x(2)-x(1))/(x(3)-x(2)))-cm(3)*(x(2)-x(1))/(x(3)-x(2))
      ELSE IF(ic1.EQ.1) THEN
        cm(1)=cm1
      ELSE IF(ic1.EQ.2) THEN
        cm(1)=-six*s1/(x(2)-x(1))-cm(2)/2.D0
      ELSE
        cm(1)=zero
      END IF

      IF(icn.EQ.0) THEN
        cm(n)=cm(n-1)*(1.D0+(x(n)-x(n-1))/(x(n-1)-x(n-2)))-cm(n-2)*(x(n)-x(n-1))/(x(n-1)-x(n-2))
      ELSE IF(icn.EQ.1) THEN
        cm(n)=cmn
      ELSE IF(icn.EQ.2) THEN
        cm(n)=-six*sn/(x(n)-x(n-1))-cm(n-1)/2.D0
      ELSE
        cm(n)=zero
      END IF

      k=2
      RETURN
      END SUBROUTINE

!************************ S P L C O F I I ****************************

      SUBROUTINE SPLCOFII(x,y,x0,y0,ia,ja,idx,idy)

!     This routine assumes that x0 and y0 are in the correct range.
!     If not, no message is issued and the results are wrong.
!     IA and JA indicate the position of X0 and y0 in the grid. They
!     give the index corresponding to the left/bottom corner of the
!     rectangle in which the point (x0,y0) is located.

      IMPLICIT REAL*8 (a-h,o-z)

      DIMENSION x(idx),y(idy)

      COMMON/UBI/ax0,bx0,cx0,dx0,ay0,by0,cy0,dy0,ax0p,bx0p,cx0p,dx0p,ay0p,by0p,cy0p,dy0p

      dxia2s6=(x(ia+1)-x(ia))**2/6.D0
      dyja2s6=(y(ja+1)-y(ja))**2/6.D0
      ax0=(x(ia+1)-x0)/(x(ia+1)-x(ia))
      bx0=1.d0-ax0
      cx0=(ax0**3-ax0)*dxia2s6
      dx0=(bx0**3-bx0)*dxia2s6

      ay0=(y(ja+1)-y0)/(y(ja+1)-y(ja))
      by0=1.d0-ay0
      cy0=(ay0**3-ay0)*dyja2s6
      dy0=(by0**3-by0)*dyja2s6

      ax0p=-1.d0/(x(ia+1)-x(ia))
      bx0p=-ax0p
      cx0p=(3.d0*ax0**2*ax0p-ax0p)*dxia2s6
      dx0p=(3.d0*bx0**2*bx0p-bx0p)*dxia2s6

      ay0p=-1.d0/(y(ja+1)-y(ja))
      by0p=-ay0p
      cy0p=(3.d0*ay0**2*ay0p-ay0p)*dyja2s6
      dy0p=(3.d0*by0**2*by0p-by0p)*dyja2s6

      END SUBROUTINE
!**************************** S P L I N 2 D *************************
      SUBROUTINE SPLIN2D(x,m,y,n,f,cx,cy,cint,idx,idy)
!
!     Two dimensional interpolation by Spline method.
!
!       Author: A. Salin     Version: 1.1    15/04/98 - 18/1/99

!     X,M: set of m abcissae
!     Y,N: set of n ordinates
!     F(i,j): value of functions at (x_i,y_j)
!     CX, CY, CINT: arrays of dimension (idx,idy) determined by
!                      subroutine SPLIN2D
!     IDX, IDY: dimension parameters in calling program

      IMPLICIT REAL*8 (a-h,o-z)
      PARAMETER (idn=999)

      DIMENSION x(idx),y(idy),f(idx,idy),cx(idx,idy),cy(idx,idy),cint(idx,idy)
      DIMENSION a(idn),aa(idn),aaa(idn),cg(idn),g(idn)

      IF(idn.LT.idx.OR.idn.LT.idy) THEN
        WRITE(*,*) 'SPLIN2D: idn too small'
        STOP
      END IF

      DO j=1,n
        CALL DSPLIN(m,x,f(1,j),cx(1,j),0.D0,0,0.D0,0,a,aa,aaa)
      END DO

      DO i=1,m
        DO j=1,n
          g(j)=f(i,j)
        END DO
        CALL DSPLIN(n,y,g,cg,0.D0,0,0.D0,0,a,aa,aaa)
        DO j=1,n
          cy(i,j)=cg(j)
        END DO
      END DO

      DO j=1,n
        CALL DSPLIN(m,x,cy(1,j),cint(1,j),0.D0,0,0.D0,0,a,aa,aaa)
      END DO

      END SUBROUTINE

!************************ S P L I N I I ******************************
      SUBROUTINE SPLINII(f,d2fdx,d2fdy,d4fdydx,ia,ja,vint,idx,idy)
!     ------------------------------------------------------------------

!     IA and JA indicate the position of X0 and y0 in the grid. They
!     give the index corresponding to the left/bottom corner of the
!     rectangle in which the point (x0,y0) is located.

      IMPLICIT REAL*8 (a-h,o-z)
      DIMENSION f(idx,idy),vint(3)
      DIMENSION d2fdx(idx,idy),d2fdy(idx,idy),d4fdydx(idx,idy)

      COMMON/UBI/ax0,bx0,cx0,dx0,ay0,by0,cy0,dy0,ax0p,bx0p,cx0p,dx0p,ay0p,by0p,cy0p,dy0p

      vint(1)= ax0*( ay0*f(ia,ja)+by0*f(ia,ja+1) &
                   + cy0*d2fdy(ia,ja)+dy0*d2fdy(ia,ja+1) ) &
             + bx0*( ay0*f(ia+1,ja)+by0*f(ia+1,ja+1) &
                   + cy0*d2fdy(ia+1,ja)+dy0*d2fdy(ia+1,ja+1) ) &
             + cx0*( ay0*d2fdx(ia,ja)+by0*d2fdx(ia,ja+1) &
                   + cy0*d4fdydx(ia,ja)+dy0*d4fdydx(ia,ja+1) ) &
             + dx0*( ay0*d2fdx(ia+1,ja)+by0*d2fdx(ia+1,ja+1) &
                   + cy0*d4fdydx(ia+1,ja)+dy0*d4fdydx(ia+1,ja+1) )
      vint(2)= ay0p*( ax0*f(ia,ja)+bx0*f(ia+1,ja) &
                    + cx0*d2fdx(ia,ja)+dx0*d2fdx(ia+1,ja) ) &
             + by0p*( ax0*f(ia,ja+1)+bx0*f(ia+1,ja+1) &
                    + cx0*d2fdx(ia,ja+1)+dx0*d2fdx(ia+1,ja+1) ) &
             + cy0p*( ax0*d2fdy(ia,ja)+bx0*d2fdy(ia+1,ja) &
                    + cx0*d4fdydx(ia,ja)+dx0*d4fdydx(ia+1,ja) ) &
             + dy0p*( ax0*d2fdy(ia,ja+1)+bx0*d2fdy(ia+1,ja+1) &
                    + cx0*d4fdydx(ia,ja+1)+dx0*d4fdydx(ia+1,ja+1) )
      vint(3)= ax0p*( ay0*f(ia,ja)+by0*f(ia,ja+1) &
                    + cy0*d2fdy(ia,ja)+dy0*d2fdy(ia,ja+1) ) &
             + bx0p*( ay0*f(ia+1,ja)+by0*f(ia+1,ja+1) &
                    + cy0*d2fdy(ia+1,ja)+dy0*d2fdy(ia+1,ja+1) ) &
             + cx0p*( ay0*d2fdx(ia,ja)+by0*d2fdx(ia,ja+1) &
                    + cy0*d4fdydx(ia,ja)+dy0*d4fdydx(ia,ja+1) ) &
             + dx0p*( ay0*d2fdx(ia+1,ja)+by0*d2fdx(ia+1,ja+1) &
                    + cy0*d4fdydx(ia+1,ja)+dy0*d4fdydx(ia+1,ja+1) )

      END SUBROUTINE
!************************ S P L I N T ********************************
!     Calculates a function Y and its derivative YP at X by spline
!     interpolation. The N data are XA (nodes) and YA (function at
!     nodes). CYA is calculated by DSPLIN
!*********************************************************************
  SUBROUTINE SPLINT(xa,ya,cya,n,x,y,yp)
    IMPLICIT NONE

    DOUBLE PRECISION :: xa(n), ya(n), cya(n)
    DOUBLE PRECISION :: x, y, yp, h, a, b
    INTEGER          :: klo, khi, k, n

    klo = 1
    khi = n

1   CONTINUE
    IF (khi - klo .GT. 1) THEN
      k = (khi + klo) / 2
      IF (xa(k) .GT. x) THEN
        khi = k
      ELSE
        klo = k
      END IF
      GO TO 1
    END IF

    h = xa(khi) - xa(klo)
    a = (xa(khi) - x) / h
    b = (x - xa(klo)) / h
    y = a * ya(klo) + b * ya(khi) + ((a**3 - a) * cya(klo) + (b**3 - b) * cya(khi)) * (h * h) / 6.D0
    yp = (ya(khi) - ya(klo)) / h + h * (cya(khi) * (3.D0 * b * b - 1.D0) - cya(klo) * (3.D0 * a * a - 1.D0)) / 6.D0
  END SUBROUTINE

! *********************** T R I C Y C L E ***********************
!     Solution of tridiagonal cyclic system
!     Required subroutine: TRIDIA
!
!         A. SALIN    Version 1    9/8/98
!
!     ALPHA: main diagonal (destroyed in TRICYCLE)
!     BETA: subdiagonal
!     GAMMA: superdiagonal
!     CBOT: element of bottom left corner of matrix
!     CTOP: element of top right corner of matrix
!     B: righthand side (destroyed in TRICYCLE)
!     X: solution
!     N: dimension of system
!     Z: scratch vector of dimension N at least
! ***************************************************************
  SUBROUTINE TRICYCLE(alpha,beta,gamma,cbot,ctop,b,x,n,z)
    IMPLICIT NONE

    DOUBLE PRECISION :: cbot, ctop, alp, fact
    DOUBLE PRECISION :: alpha(n), beta(n), gamma(n), b(n), x(n), z(n)
    INTEGER          :: i, n

    alp = -alpha(1)
    alpha(1) = alpha(1) - alp
    alpha(n) = alpha(n) - cbot * ctop / alp
    DO i = 1, n
      z(i) = alpha(i)
    END DO
    CALL TRIDIA( alpha, beta, gamma, b, x, n )

    DO i = 1, n
      alpha(i) = z(i)
    END DO
    b(1) = alp
    b(n) = cbot
    DO i = 2, n - 1
      b(i) = 0.D0
    END DO
    CALL TRIDIA( alpha, beta, gamma, b, z, n )
    fact = (x(1) + ctop * x(n) / alp) / (1.D0 + z(1) + ctop * z(n) / alp)
    DO i = 1, n
      x(i) = x(i) - fact * z(i)
    END DO
  END SUBROUTINE

!**************************** T R I D I A ***************************
!
!        Solution of tridiagonal systems
!
!     ALPHA: main diagonal (destroyed in TRIDIA).
!     BETA: subdiagonal (first element has index 1).
!     GAMMA: superdiagonal (first element has index 1).
!     B: right-hand side (destroyed in TRIDIA).
!     X: solution.
!     N: dimension of system.
!********************************************************************
  SUBROUTINE TRIDIA( alpha, beta, gamma, b, x, n )
    IMPLICIT NONE

    DOUBLE PRECISION :: rap
    DOUBLE PRECISION :: alpha(n), beta(n), gamma(n), b(n), x(n)
    INTEGER          :: i, j, n

    DO i = 2, n
      rap = beta(i - 1) / alpha(i - 1)
      alpha(i) = alpha(i) - rap * gamma(i - 1)
      b(i) = b(i) - rap * b(i - 1)
    END DO
    x(n) = b(n) / alpha(n)

    DO j = 2, n
      i = n - j + 1
      x(i) = ( b(i) - gamma(i) * x(i + 1) ) / alpha(i)
    END DO
  END SUBROUTINE

! ************************ T R I S P L ***************************
!     Spline interpolation of a periodic function with period
!     [x(n)-x(1)] - i.e. y(n)=y(1).

!     See general comments in DSPLIN.
!     The vectors ALPHA, BETA, GAMMA, B, BZ are scratch vectors of
!     dimension at list N.
!     Required subroutine: TRIDIA, TRICYCLE
!
!        Author: A. Salin   Version 1  09/06/98
! ***************************************************************
  SUBROUTINE TRISPL( n, x, y, cm, alpha, beta, gamma, b, bz )
    IMPLICIT NONE

    DOUBLE PRECISION :: x(n), y(n), cm(n)
    DOUBLE PRECISION :: alpha(n), beta(n), gamma(n), b(n), bz(n)

    INTEGER :: kod, k, iw, iligne
    COMMON /KODSPL/ kod, k, iw, iligne

    INTEGER :: n, i, i2
    DOUBLE PRECISION :: a, c, d, e, cbot, ctop

91  FORMAT("Error in the data for TRISPL",/,"Check abcissae:",1PD15.8,1X," and ",D15.8,/,"Program stopped")

    kod = 0
    c = x(2) - x(1)
    e = (y(2) - y(1)) / c
    DO i = 3, n
      i2 = i - 2
      a = c
      c = x(i) - x(i - 1)
      IF (a * c .LE. 0.0) THEN
        kod = -1
        IF (iw .GT. 0) WRITE(iw, 91) x(i - 1), x(i)
        STOP
      END IF
      alpha(i2) = (a + c) / 3.0
      beta(i2) = c / 6.0
      gamma(i2) = beta(i2)
      d = e
      e = (y(i) - y(i - 1)) / c
      b(i2) = e - d
    END DO

    a = c
    c = x(2) - x(1)
    alpha(n - 1) = (a + c) / 3.0
    beta(n - 1) = c / 6.0
    gamma(n - 1) = beta(n - 1)
    d = e
    e = (y(2) - y(1)) / c
    b(n - 1) = e - d
    cbot = beta(n - 1)
    ctop = cbot

!   Solve tridiagonal system
    CALL TRICYCLE( alpha, beta, gamma, cbot, ctop, b, cm(2), n - 1, bz )

    cm(1) = cm(n)

    k = 2
  END SUBROUTINE

!**************************** L O C A T E ***************************
!       Author: A. Salin     Version: 1.0    15/04/98
!
!     Determines k such that t.GT.x(k-1) and t.LE.x(k)
!     (except for t=x(1) in which case k=2)
!     If k is initially in the interval [1,n], then the search starts
!     from x(k).
!********************************************************************
  SUBROUTINE LOCATE( t, k, x, n)
    IMPLICIT NONE

    DOUBLE PRECISION :: t, x(n)
    INTEGER          :: k, n, inc, klo, khi, km
    LOGICAL          :: order

    order = x(2) .GT. x(1)
    IF (k .LE. 1 .OR. k .GT. n) THEN
      klo = 0
      khi = n + 1
      GO TO 100
    END IF

    inc = 1
    klo = k - 1
    IF (t .GT. x(klo) .EQV. order) THEN
 10   khi = klo + inc
      IF (khi .GT. n) THEN
        khi = n + 1
      ELSE IF (t .GT. x(khi) .EQV. order) THEN
        klo = khi
        inc = inc + inc
        GO TO 10
      END IF
    ELSE
      khi = klo
 20   klo = khi - inc
      IF (klo .LT. 1) THEN
        klo = 0
      ELSE IF (t .LT. x(klo) .EQV. order) THEN
        khi = klo
        inc = inc + inc
        GO TO 20
      END IF
    END IF

100 CONTINUE
    IF (khi - klo .GT. 1) THEN
      km = (khi + klo) / 2
      IF (t .GT. x(km) .EQV. order) THEN
        klo = km
      ELSE
        khi = km
      END IF
      GO TO 100
    END IF

    k = klo + 1

    IF (t .EQ. x(1)) k = 2
  END SUBROUTINE

!************************ U B I C A R ********************************
!   Determine KLO such that X is .GE.XA(KLO) and .LT.XA(KLO+1)
!   If X is outside the interval [XA(1),XA(N)], the logical
!   variable PARO takes the value .TRUE.
!*********************************************************************
  SUBROUTINE UBICAR(XA,X,N,KLO,PARO)
    IMPLICIT NONE

    DOUBLE PRECISION :: x, xa(N)
    INTEGER :: n, klo, khi, k
    LOGICAL PARO

    paro = .FALSE.
    IF (x .LT. xa(1) .OR. x .GT. xa(n)) THEN
      PARO = .TRUE.
      RETURN
    END IF
    klo = 1
    khi = n
1   CONTINUE
    IF (khi - klo .GT. 1) THEN
      k = (khi + klo) / 2
      IF (xa(k) .GT. x) THEN
        khi = k
      ELSE
        klo = k
      END IF
      GO TO 1
    END IF
  END SUBROUTINE
END MODULE

